/*
** 
** Mapinfo conversion tool (based on ZDoom's old MAPINFO parser)
**
**---------------------------------------------------------------------------
** Copyright 1998-2008 Randy Heit
** Copyright 2009 Christoph Oelckers
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions
** are met:
**
** 1. Redistributions of source code must retain the above copyright
**    notice, this list of conditions and the following disclaimer.
** 2. Redistributions in binary form must reproduce the above copyright
**    notice, this list of conditions and the following disclaimer in the
**    documentation and/or other materials provided with the distribution.
** 3. The name of the author may not be used to endorse or promote products
**    derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
** IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
** OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
** IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
** INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
** NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
** DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
** THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
** THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**---------------------------------------------------------------------------
**
*/

#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "sc_man.h"

#ifdef _MSC_VER
#pragma warning(disable:4996)
#endif

static void ParseEpisodeInfo (FILE *out);
static void ParseSkill (FILE *out);

static bool HexenHack;


static const char *MapInfoTopLevel[] =
{
	"map",
	"defaultmap",
	"clusterdef",
	"episode",
	"clearepisodes",
	"skill",
	"clearskills",
	"adddefaultmap",
	"gamedefaults",
	NULL
};

enum
{
	MITL_MAP,
	MITL_DEFAULTMAP,
	MITL_CLUSTERDEF,
	MITL_EPISODE,
	MITL_CLEAREPISODES,
	MITL_SKILL,
	MITL_CLEARSKILLS,
	MITL_ADDDEFAULTMAP,
	MITL_GAMEDEFAULTS,
};

static const char *MapInfoMapLevel[] =
{
	"levelnum",
	"next",
	"secretnext",
	"cluster",
	"sky1",
	"sky2",
	"fade",
	"outsidefog",
	"titlepatch",
	"par",
	"sucktime",
	"music",
	"nointermission",
 	"intermission",
	"doublesky",
	"nosoundclipping",
	"allowmonstertelefrags",
	"map07special",
	"baronspecial",
	"cyberdemonspecial",
	"spidermastermindspecial",
	"minotaurspecial",
	"dsparilspecial",
	"ironlichspecial",
	"specialaction_exitlevel",
	"specialaction_opendoor",
	"specialaction_lowerfloor",
	"specialaction_killmonsters",
	"lightning",
	"fadetable",
	"evenlighting",
	"smoothlighting",
	"noautosequences",
	"autosequences",
	"forcenoskystretch",
	"skystretch",
	"allowfreelook",
	"nofreelook",
	"allowjump",
	"nojump",
	"fallingdamage",		// Hexen falling damage
	"oldfallingdamage",		// Lesser ZDoom falling damage
	"forcefallingdamage",	// Skull Tag compatibility name for oldfallingdamage
	"strifefallingdamage",	// Strife's falling damage is really unforgiving
	"nofallingdamage",
	"noallies",
	"cdtrack",
	"cdid",
	"cd_start_track",
	"cd_end1_track",
	"cd_end2_track",
	"cd_end3_track",
	"cd_intermission_track",
	"cd_title_track",
	"warptrans",
	"vertwallshade",
	"horizwallshade",
	"gravity",
	"aircontrol",
	"filterstarts",
	"activateowndeathspecials",
	"killeractivatesdeathspecials",
	"missilesactivateimpactlines",
	"missileshootersactivetimpactlines",
	"noinventorybar",
	"deathslideshow",
	"redirect",
	"strictmonsteractivation",
	"laxmonsteractivation",
	"additive_scrollers",
	"interpic",
	"exitpic",
	"enterpic",
	"intermusic",
	"airsupply",
	"specialaction",
	"keepfullinventory",
	"monsterfallingdamage",
	"nomonsterfallingdamage",
	"sndseq",
	"sndinfo",
	"soundinfo",
	"clipmidtextures",
	"wrapmidtextures",
	"allowcrouch",
	"nocrouch",
	"pausemusicinmenus",
	"compat_shorttex",	
	"compat_stairs",		
	"compat_limitpain",	
	"compat_nopassover",	
	"compat_notossdrops",	
	"compat_useblocking", 
	"compat_nodoorlight",	
	"compat_ravenscroll",	
	"compat_soundtarget",	
	"compat_dehhealth",	
	"compat_trace",		
	"compat_dropoff",
	"compat_boomscroll",
	"compat_invisibility",
	"compat_silent_instant_floors",
	"compat_sectorsounds",
	"compat_missileclip",
	"bordertexture",
	"f1", // [RC] F1 help
	"noinfighting",
	"normalinfighting",
	"totalinfighting",
	"infiniteflightpowerup",
	"noinfiniteflightpowerup",
	"allowrespawn",
	"teamdamage",
	"teamplayon",
	"teamplayoff",
	"checkswitchrange",
	"nocheckswitchrange",
	"translator",
	"unfreezesingleplayerconversations",
	"nobotnodes",
	NULL
};

enum EMIType
{
	MITYPE_EATNEXT,
	MITYPE_IGNORE,
	MITYPE_INT,
	MITYPE_FLOAT,
	MITYPE_HEX,
	MITYPE_COLOR,
	MITYPE_MAPNAME,
	MITYPE_LUMPNAME,
	MITYPE_SKY,
	MITYPE_SETFLAG,
	MITYPE_CLRFLAG,
	MITYPE_SCFLAGS,
	MITYPE_CLUSTER,
	MITYPE_STRING,
	MITYPE_MUSIC,
	MITYPE_RELLIGHT,
	MITYPE_CLRBYTES,
	MITYPE_REDIRECT,
	MITYPE_SPECIALACTION,
	MITYPE_COMPATFLAG,
	MITYPE_STRINGT,
};

EMIType MapHandlers[] =
{
	 MITYPE_INT,		
	 MITYPE_MAPNAME,	
	 MITYPE_MAPNAME,	
	 MITYPE_CLUSTER,	
	 MITYPE_SKY,		
	 MITYPE_SKY,		
	 MITYPE_COLOR,		
	 MITYPE_COLOR,		
	 MITYPE_LUMPNAME,	
	 MITYPE_INT,		
	 MITYPE_INT,		
	 MITYPE_MUSIC,		
	 MITYPE_SETFLAG,	
	 MITYPE_CLRFLAG,	
	 MITYPE_SETFLAG,	
	 MITYPE_IGNORE,	
	 MITYPE_SETFLAG,	
	 MITYPE_SETFLAG,	
	 MITYPE_SETFLAG,	
	 MITYPE_SETFLAG,	
	 MITYPE_SETFLAG,	
	 MITYPE_SETFLAG,	
	 MITYPE_SETFLAG,	
	 MITYPE_SETFLAG,	
	 MITYPE_SCFLAGS,	
	 MITYPE_SCFLAGS,	
	 MITYPE_SCFLAGS,	
	 MITYPE_SETFLAG,	
	 MITYPE_SETFLAG,	
	 MITYPE_LUMPNAME,	
	 MITYPE_CLRBYTES,	
	 MITYPE_SETFLAG,	
	 MITYPE_SETFLAG,	
	 MITYPE_CLRFLAG,	
	 MITYPE_SETFLAG,	
	 MITYPE_CLRFLAG,	
	 MITYPE_SCFLAGS,	
	 MITYPE_SCFLAGS,	
	 MITYPE_CLRFLAG,	
	 MITYPE_SETFLAG,	
	 MITYPE_SCFLAGS,	
	 MITYPE_SCFLAGS,	
	 MITYPE_SCFLAGS,	
	 MITYPE_SETFLAG,	
	 MITYPE_SCFLAGS,	
	 MITYPE_SETFLAG,	
	 MITYPE_INT,		
	 MITYPE_HEX,		
	 MITYPE_EATNEXT,	
	 MITYPE_EATNEXT,	
	 MITYPE_EATNEXT,	
	 MITYPE_EATNEXT,	
	 MITYPE_EATNEXT,	
	 MITYPE_EATNEXT,	
	 MITYPE_INT,		
	 MITYPE_RELLIGHT,	
	 MITYPE_RELLIGHT,	
	 MITYPE_FLOAT,		
	 MITYPE_FLOAT,		
	 MITYPE_SETFLAG,	
	 MITYPE_SETFLAG,	
	 MITYPE_CLRFLAG,	
	 MITYPE_SETFLAG,	
	 MITYPE_CLRFLAG,	
	 MITYPE_SETFLAG,	
	 MITYPE_SETFLAG,	
	 MITYPE_REDIRECT,	
	 MITYPE_CLRFLAG,	
	 MITYPE_SETFLAG,	
	 MITYPE_COMPATFLAG,
	 MITYPE_STRING,	
	 MITYPE_STRING,	
	 MITYPE_STRING,	
	 MITYPE_MUSIC,		
	 MITYPE_INT,		
	 MITYPE_SPECIALACTION,
	 MITYPE_SETFLAG,	
	 MITYPE_SETFLAG,	
	 MITYPE_CLRFLAG,	
	 MITYPE_STRING,	
	 MITYPE_STRING,	
	 MITYPE_STRING,	
	 MITYPE_SETFLAG,	
	 MITYPE_SETFLAG,	
	 MITYPE_CLRFLAG,	
	 MITYPE_SETFLAG,	
	 MITYPE_SCFLAGS,	
	 MITYPE_COMPATFLAG,
	 MITYPE_COMPATFLAG,
	 MITYPE_COMPATFLAG,
	 MITYPE_COMPATFLAG,
	 MITYPE_COMPATFLAG,
	 MITYPE_COMPATFLAG,
	 MITYPE_COMPATFLAG,
	 MITYPE_COMPATFLAG,
	 MITYPE_COMPATFLAG,
	 MITYPE_COMPATFLAG,
	 MITYPE_COMPATFLAG,
	 MITYPE_COMPATFLAG,
	 MITYPE_COMPATFLAG,
	 MITYPE_COMPATFLAG,
	 MITYPE_COMPATFLAG,
	 MITYPE_COMPATFLAG,
	 MITYPE_COMPATFLAG,
	 MITYPE_LUMPNAME,	
	 MITYPE_LUMPNAME,  
	 MITYPE_SCFLAGS,	
	 MITYPE_SCFLAGS,	
	 MITYPE_SCFLAGS,	
	 MITYPE_SETFLAG,	
	 MITYPE_CLRFLAG,	
	 MITYPE_SETFLAG,	
	 MITYPE_FLOAT,		
	 MITYPE_SCFLAGS,	
	 MITYPE_SCFLAGS,	
	 MITYPE_SETFLAG,	
	 MITYPE_CLRFLAG,	
	 MITYPE_STRING,	
	 MITYPE_SETFLAG,	
	 MITYPE_IGNORE,	
};

static const char *MapInfoClusterLevel[] =
{
	"entertext",
	"exittext",
	"music",
	"flat",
	"pic",
	"hub",
	"cdtrack",
	"cdid",
	"entertextislump",
	"exittextislump",
	"name",
	NULL
};

EMIType ClusterHandlers[] =
{
	 MITYPE_STRINGT,	
	 MITYPE_STRINGT,	
	 MITYPE_MUSIC,		
	 MITYPE_LUMPNAME,	
	 MITYPE_LUMPNAME,	
	 MITYPE_SETFLAG,	
	 MITYPE_INT,		
	 MITYPE_HEX,		
	 MITYPE_SETFLAG,	
	 MITYPE_SETFLAG,	
	 MITYPE_STRINGT,	
};

static void ParseMapInfoLower (FILE *out,
							   EMIType *handlers,
							   const char *strings[]);

// [RH] Returns true if the specified string is a valid decimal number

bool IsNum (const char *str)
{
	while (*str)
	{
		if (((*str < '0') || (*str > '9')) && (*str != '-'))
		{
			return false;
		}
		str++;
	}
	return true;
}

int main(int argc, const char **argv)
{
	HexenHack = false;

	if (argc < 3)
	{
		printf("Mapinfo converter\nUsage: mapinfoconv <infile> <outfile>\n");
		exit(1);
	}

	SC_OpenFile(argv[1]);

	FILE *out = fopen(argv[2], "wt");
	if (out == NULL)
	{
		printf("%s: Unable to create file\n", argv[2]);
		exit(1);
	}
	sc_Out = out;

	while (SC_GetString ())
	{
		switch (SC_MustMatchString (MapInfoTopLevel))
		{
		case MITL_GAMEDEFAULTS:
		case MITL_DEFAULTMAP:
		case MITL_ADDDEFAULTMAP:
			fprintf(out, "%s\n{\n", sc_String);
			ParseMapInfoLower (out, MapHandlers, MapInfoMapLevel);
			break;

		case MITL_MAP:		// map <MAPNAME> <Nice Name>
			SC_MustGetString ();
			fprintf(out, "map %s ", sc_String);
			if (IsNum (sc_String))
			{	
				HexenHack = true;
			}
			SC_MustGetString ();
			if (SC_Compare("lookup"))
			{
				fprintf(out, "lookup ");
				SC_MustGetString();
			}
			fprintf(out, "%s\n{\n", sc_String);

			ParseMapInfoLower (out, MapHandlers, MapInfoMapLevel);
			break;

		case MITL_CLUSTERDEF:	// clusterdef <clusternum>
			SC_MustGetNumber ();
			fprintf(out, "cluster %d\n{\n", sc_Number);
			ParseMapInfoLower (out, ClusterHandlers, MapInfoClusterLevel);
			break;

		case MITL_EPISODE:
			ParseEpisodeInfo(out);
			break;

		case MITL_CLEAREPISODES:
			fprintf(out, "clearepisodes\n");
			break;

		case MITL_SKILL:
			ParseSkill(out);
			break;

		case MITL_CLEARSKILLS:
			fprintf(out, "clearskills\n");
			break;

		default:
			SC_ScriptError("Unknown keyword %s\n", sc_String);
			break;

		}
	}
	fclose(out);
}

static void ParseMapInfoLower (FILE *out,
							   EMIType *handlers,
							   const char *strings[])
{
	int entry;
	EMIType *handler;

	while (SC_GetString ())
	{
		if (SC_MatchString (MapInfoTopLevel) != -1)
		{
			SC_UnGet ();
			break;
		}
		entry = SC_MatchString (strings);

		fprintf(out, "\t%s ", sc_String);

		handler = handlers + entry;
		switch (*handler)
		{
		case MITYPE_STRING:
		case MITYPE_EATNEXT:
		case MITYPE_LUMPNAME:
		case MITYPE_COLOR:
		case MITYPE_HEX:
		case MITYPE_MUSIC:
			SC_MustGetString ();
			fprintf(out, "= %s\n", sc_String);
			break;

		case MITYPE_IGNORE:
		case MITYPE_SETFLAG:
		case MITYPE_CLRFLAG:
		case MITYPE_SCFLAGS:
		case MITYPE_CLRBYTES:
			fprintf(out, "\n");
			break;

		case MITYPE_INT:
		case MITYPE_CLUSTER:
		case MITYPE_RELLIGHT:
			SC_MustGetNumber ();
			fprintf(out, "= %s\n", sc_String);	// output the actual string that was parsed
			break;

		case MITYPE_FLOAT:
			SC_MustGetFloat ();
			fprintf(out, "= %s\n", sc_String);	// output the actual string that was parsed
			break;

		case MITYPE_REDIRECT:
			SC_MustGetString ();
			fprintf(out, "= %s\n", sc_String);
			SC_MustGetString ();
			fprintf(out, ", %s\n", sc_String);
			break;

		case MITYPE_MAPNAME:
			SC_MustGetString ();

			if (SC_Compare("endgame"))
			{
				SC_MustGetStringName("{");
				fprintf(out, "= endgame\n\t{\n");
				while (!SC_CheckString("}"))
				{
					SC_MustGetString();
					if (SC_Compare("pic"))
					{
						SC_MustGetString();
						fprintf(out, "\t\tpic = %s\n", sc_String);
					}
					else if (SC_Compare("hscroll"))
					{
						SC_MustGetString();
						fprintf(out, "\t\thscroll = %s", sc_String);
						SC_MustGetString();
						fprintf(out, ", %s", sc_String);
						if (SC_CheckNumber())
						{
							fprintf(out, ", %s", sc_String);
						}
						fputc('\n', out);
					}
					else if (SC_Compare("vscroll"))
					{
						SC_MustGetString();
						fprintf(out, "\t\thscroll = %s", sc_String);
						SC_MustGetString();
						fprintf(out, ", %s\n", sc_String);
					}
					else if (SC_Compare("cast"))
					{
						fprintf(out, "\t\tcast\n");
					}
					else if (SC_Compare("music"))
					{
						SC_MustGetString();
						fprintf(out, "\t\tmusic = %s", sc_String);
						if (SC_CheckNumber())
						{
							fprintf(out, ", %s", sc_String);
						}
						fputc('\n', out);
					}
				}
				fprintf(out, "\t}\n");
			}
			else if (SC_Compare("endpic"))
			{
				SC_MustGetString ();
				fprintf(out, "= endpic, %s\n", sc_String);
			}
			else if (!IsNum(sc_String))
			{
				fprintf(out, "= %s\n", sc_String);
			}
			else
			{
				fprintf(out, ", %s\n", sc_String);
			}
			break;

		case MITYPE_SKY:
			SC_MustGetString ();	// get texture name;
			fprintf(out, "= %s", sc_String);
			SC_MustGetFloat ();		// get scroll speed
			fprintf(out, ", %s\n", sc_String);
			break;

		case MITYPE_STRINGT:
			if (!SC_GetString (true))
			{
				SC_ScriptError ("Missing string (unexpected end of file).");
			}
			if (SC_Compare("lookup"))
			{
				SC_MustGetString ();
				fprintf(out, "= lookup, %s\n", sc_String);
			}
			else
			{
				fprintf(out, "= %s\n", sc_String);
			}
			break;

		case MITYPE_SPECIALACTION:
			{
				SC_SetCMode(true);
				SC_MustGetString();
				fprintf(out, "= %s", sc_String);
				SC_CheckString(",");
				SC_MustGetString();
				fprintf(out, ", %s", sc_String);
				int j = 0;
				while (j < 5 && SC_CheckString(","))
				{
					SC_MustGetNumber();
					fprintf(out, ", %s", sc_String);
				}
				SC_SetCMode(false);
				fputc('\n', out);
			}
			break;

		case MITYPE_COMPATFLAG:
			if (!SC_CheckNumber()) fputc('\n', out);
			else
			{
				fprintf(out, "= %s\n", sc_String);
			}
			break;
		}
	}
	fprintf(out,"}\n\n");
}

// Episode definitions start with the header "episode <start-map>"
// and then can be followed by any of the following:
//
// name "Episode name as text"
// picname "Picture to display the episode name"
// key "Shortcut key for the menu"
// noskillmenu
// remove

static void ParseEpisodeInfo (FILE *out)
{
	// Get map name
	SC_MustGetString ();
	fprintf(out, "episode %s", sc_String);

	SC_MustGetString ();
	if (SC_Compare ("teaser"))
	{
		SC_MustGetString ();
		fprintf(out, " teaser %s", sc_String);
		SC_MustGetString ();
	}
	fprintf(out, "\n{\n");
	do
	{
		if (SC_Compare ("optional") ||
			SC_Compare ("extended") ||
			SC_Compare ("remove") ||
			SC_Compare ("noskillmenu"))
		{
			fprintf(out, "\t%s\n", sc_String);
		}
		else if (SC_Compare ("name") ||
			SC_Compare ("picname") ||
			SC_Compare ("key"))
		{
			fprintf(out, "\t%s = ", sc_String);
			SC_MustGetString ();
			fprintf(out, "%s\n", sc_String);
		}
		else
		{
			SC_UnGet ();
			break;
		}
	}
	while (SC_GetString ());
	fprintf(out,"}\n\n");
}



static void ParseSkill (FILE *out)
{
	SC_MustGetString();
	fprintf(out, "skill %s\n{\n", sc_String);

	while (SC_GetString ())
	{
		if (SC_Compare ("ammofactor") ||
			SC_Compare ("doubleammofactor") ||
			SC_Compare ("dropammofactor") ||
			SC_Compare ("damagefactor") ||
			SC_Compare("respawntime") ||
			SC_Compare("Aggressiveness"))
		{
			fprintf(out, "\t%s = ", sc_String);
			SC_MustGetFloat ();
			fprintf(out, "%s\n", sc_String);
		}
		else if (SC_Compare ("fastmonsters") ||
			SC_Compare ("disablecheats") ||
			SC_Compare ("easybossbrain") ||
			SC_Compare("autousehealth"))
		{
			fprintf(out, "\t%s\n", sc_String);
		}
		else if (SC_Compare("respawnlimit") ||
			SC_Compare("ACSReturn"))
		{
			fprintf(out, "\t%s = ", sc_String);
			SC_MustGetNumber ();
			fprintf(out, "%s\n", sc_String);
		}
		else if (SC_Compare("SpawnFilter"))
		{
			fprintf(out, "\t%s = ", sc_String);
			SC_MustGetString ();
			fprintf(out, "%s\n", sc_String);
		}
		else if (SC_Compare("Name") ||
			SC_Compare("PicName") ||
			SC_Compare("Key") ||
			SC_Compare("TextColor"))
		{
			fprintf(out, "\t%s = ", sc_String);
			SC_MustGetString ();
			fprintf(out, "%s\n", sc_String);
		}
		else if (SC_Compare("MustConfirm"))
		{
			fprintf(out, "\t%s", sc_String);
			if (SC_GetString() && sc_StringQuoted)
			{
			}
			else
			{
				SC_UnGet();
			}
		}
		else if (SC_Compare("PlayerClassName"))
		{
			SC_MustGetString ();
			fprintf(out, "\tplayerclassname = %s, ", sc_String);
			SC_MustGetString ();
			fprintf(out, "%s\n", sc_String);
		}
		else
		{
			SC_UnGet ();
			break;
		}
	}
	fprintf(out,"}\n\n");
}
